@@ -9,7 +9,7 @@
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2004 Julian Seward 
+   Copyright (C) 2000-2005 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -30,8 +30,8 @@
    The GNU General Public License is contained in the file COPYING.
 */
 
-#include "vg_include.h"
-#include "vg_arch_syscall.h"
+#include "core.h"
+#include <stdio.h>
 
 /* ---------------------------------------------------------------------
    Wrappers around system calls, and other stuff, to do with signals.
@@ -40,75 +40,75 @@
 /* sigemptyset, sigfullset, sigaddset and sigdelset return 0 on
    success and -1 on error.  
 */
-Int VG_(ksigfillset)( vki_ksigset_t* set )
+Int VG_(sigfillset)( vki_sigset_t* set )
 {
    Int i;
    if (set == NULL)
       return -1;
-   for (i = 0; i < VKI_KNSIG_WORDS; i++)
-      set->ws[i] = 0xFFFFFFFF;
+   for (i = 0; i < _VKI_NSIG_WORDS; i++)
+      set->sig[i] = ~(UWord)0x0;
    return 0;
 }
 
-Int VG_(ksigemptyset)( vki_ksigset_t* set )
+Int VG_(sigemptyset)( vki_sigset_t* set )
 {
    Int i;
    if (set == NULL)
       return -1;
-   for (i = 0; i < VKI_KNSIG_WORDS; i++)
-      set->ws[i] = 0x0;
+   for (i = 0; i < _VKI_NSIG_WORDS; i++)
+      set->sig[i] = 0x0;
    return 0;
 }
 
-Bool VG_(kisemptysigset)( vki_ksigset_t* set )
+Bool VG_(isemptysigset)( const vki_sigset_t* set )
 {
    Int i;
    vg_assert(set != NULL);
-   for (i = 0; i < VKI_KNSIG_WORDS; i++)
-      if (set->ws[i] != 0x0) return False;
+   for (i = 0; i < _VKI_NSIG_WORDS; i++)
+      if (set->sig[i] != 0x0) return False;
    return True;
 }
 
-Bool VG_(kisfullsigset)( vki_ksigset_t* set )
+Bool VG_(isfullsigset)( const vki_sigset_t* set )
 {
    Int i;
    vg_assert(set != NULL);
-   for (i = 0; i < VKI_KNSIG_WORDS; i++)
-      if (set->ws[i] != (UInt)(~0x0)) return False;
+   for (i = 0; i < _VKI_NSIG_WORDS; i++)
+      if (set->sig[i] != ~(UWord)0x0) return False;
    return True;
 }
 
 
-Int VG_(ksigaddset)( vki_ksigset_t* set, Int signum )
+Int VG_(sigaddset)( vki_sigset_t* set, Int signum )
 {
    if (set == NULL)
       return -1;
-   if (signum < 1 || signum > VKI_KNSIG)
+   if (signum < 1 || signum > _VKI_NSIG)
       return -1;
    signum--;
-   set->ws[signum / VKI_KNSIG_BPW] |= (1 << (signum % VKI_KNSIG_BPW));
+   set->sig[signum / _VKI_NSIG_BPW] |= (1 << (signum % _VKI_NSIG_BPW));
    return 0;
 }
 
-Int VG_(ksigdelset)( vki_ksigset_t* set, Int signum )
+Int VG_(sigdelset)( vki_sigset_t* set, Int signum )
 {
    if (set == NULL)
       return -1;
-   if (signum < 1 || signum > VKI_KNSIG)
+   if (signum < 1 || signum > _VKI_NSIG)
       return -1;
    signum--;
-   set->ws[signum / VKI_KNSIG_BPW] &= ~(1 << (signum % VKI_KNSIG_BPW));
+   set->sig[signum / _VKI_NSIG_BPW] &= ~(1 << (signum % _VKI_NSIG_BPW));
    return 0;
 }
 
-Int VG_(ksigismember) ( vki_ksigset_t* set, Int signum )
+Int VG_(sigismember) ( const vki_sigset_t* set, Int signum )
 {
    if (set == NULL)
       return 0;
-   if (signum < 1 || signum > VKI_KNSIG)
+   if (signum < 1 || signum > _VKI_NSIG)
       return 0;
    signum--;
-   if (1 & ((set->ws[signum / VKI_KNSIG_BPW]) >> (signum % VKI_KNSIG_BPW)))
+   if (1 & ((set->sig[signum / _VKI_NSIG_BPW]) >> (signum % _VKI_NSIG_BPW)))
       return 1;
    else
       return 0;
@@ -116,97 +116,97 @@ Int VG_(ksigismember) ( vki_ksigset_t* set, Int signum )
 
 
 /* Add all signals in src to dst. */
-void VG_(ksigaddset_from_set)( vki_ksigset_t* dst, vki_ksigset_t* src )
+void VG_(sigaddset_from_set)( vki_sigset_t* dst, vki_sigset_t* src )
 {
    Int i;
    vg_assert(dst != NULL && src != NULL);
-   for (i = 0; i < VKI_KNSIG_WORDS; i++)
-      dst->ws[i] |= src->ws[i];
+   for (i = 0; i < _VKI_NSIG_WORDS; i++)
+      dst->sig[i] |= src->sig[i];
 }
 
 /* Remove all signals in src from dst. */
-void VG_(ksigdelset_from_set)( vki_ksigset_t* dst, vki_ksigset_t* src )
+void VG_(sigdelset_from_set)( vki_sigset_t* dst, vki_sigset_t* src )
 {
    Int i;
    vg_assert(dst != NULL && src != NULL);
-   for (i = 0; i < VKI_KNSIG_WORDS; i++)
-      dst->ws[i] &= ~(src->ws[i]);
+   for (i = 0; i < _VKI_NSIG_WORDS; i++)
+      dst->sig[i] &= ~(src->sig[i]);
 }
 
 
 /* The functions sigaction, sigprocmask, sigpending and sigsuspend
    return 0 on success and -1 on error.  
 */
-Int VG_(ksigprocmask)( Int how, 
-                       const vki_ksigset_t* set, 
-                       vki_ksigset_t* oldset)
+Int VG_(sigprocmask)( Int how, const vki_sigset_t* set, vki_sigset_t* oldset)
 {
    Int res 
       = VG_(do_syscall)(__NR_rt_sigprocmask, 
-			how, (UInt)set, (UInt)oldset, 
-			VKI_KNSIG_WORDS * VKI_BYTES_PER_WORD);
+			how, (UWord)set, (UWord)oldset, 
+			_VKI_NSIG_WORDS * sizeof(UWord));
    return VG_(is_kerror)(res) ? -1 : 0;
 }
 
 
-Int VG_(ksigaction) ( Int signum,  
-                      const vki_ksigaction* act,  
-                      vki_ksigaction* oldact)
+Int VG_(sigaction) ( Int signum, const struct vki_sigaction* act,  
+                     struct vki_sigaction* oldact)
 {
    Int res
      = VG_(do_syscall)(__NR_rt_sigaction,
-		       signum, (UInt)act, (UInt)oldact, 
-		       VKI_KNSIG_WORDS * VKI_BYTES_PER_WORD);
+		       signum, (UWord)act, (UWord)oldact, 
+		       _VKI_NSIG_WORDS * sizeof(UWord));
    /* VG_(printf)("res = %d\n",res); */
    return VG_(is_kerror)(res) ? -1 : 0;
 }
 
 
-Int VG_(ksigaltstack)( const vki_kstack_t* ss, vki_kstack_t* oss )
+Int VG_(sigaltstack)( const vki_stack_t* ss, vki_stack_t* oss )
 {
-   Int res
-     = VG_(do_syscall)(__NR_sigaltstack, (UInt)ss, (UInt)oss);
+   Int res = VG_(do_syscall)(__NR_sigaltstack, (UWord)ss, (UWord)oss);
    return VG_(is_kerror)(res) ? -1 : 0;
 }
 
-Int VG_(ksigtimedwait)( const vki_ksigset_t *set, vki_ksiginfo_t *info, 
+Int VG_(sigtimedwait)( const vki_sigset_t *set, vki_siginfo_t *info, 
 			const struct vki_timespec *timeout )
 {
    Int res = VG_(do_syscall)(__NR_rt_sigtimedwait, set, info, timeout, sizeof(*set));
 
-   return VG_(is_kerror)(res) ? -1 : res;
+   return res;
 }
  
-Int VG_(ksignal)(Int signum, void (*sighandler)(Int))
+Int VG_(signal)(Int signum, void (*sighandler)(Int))
 {
    Int res;
-   vki_ksigaction sa;
+   struct vki_sigaction sa;
    sa.ksa_handler = sighandler;
-   sa.ksa_flags = VKI_SA_ONSTACK | VKI_SA_RESTART;
-   sa.ksa_restorer = NULL;
-   res = VG_(ksigemptyset)( &sa.ksa_mask );
+   sa.sa_flags = VKI_SA_ONSTACK | VKI_SA_RESTART;
+   sa.sa_restorer = NULL;
+   res = VG_(sigemptyset)( &sa.sa_mask );
    vg_assert(res == 0);
    res = VG_(do_syscall)(__NR_rt_sigaction,
-			 signum, (UInt)(&sa), (UInt)NULL,
-			 VKI_KNSIG_WORDS * VKI_BYTES_PER_WORD);
+			 signum, (UWord)&sa, (UWord)NULL,
+			 _VKI_NSIG_WORDS * sizeof(UWord));
    return VG_(is_kerror)(res) ? -1 : 0;
 }
 
 
-Int VG_(kkill)( Int pid, Int signo )
+Int VG_(kill)( Int pid, Int signo )
 {
    Int res = VG_(do_syscall)(__NR_kill, pid, signo);
    return VG_(is_kerror)(res) ? -1 : 0;
 }
 
 
-Int VG_(ktkill)( Int tid, Int signo )
+Int VG_(tkill)( Int tid, Int signo )
 {
    Int ret = -VKI_ENOSYS;
 
-#ifdef __NR_tkill
+#if 0
+   /* This isn't right because the client may create a process
+      structure with multiple thread groups */
+   ret = VG_(do_syscall)(__NR_tgkill, VG_(getpid)(), tid, signo);
+#endif
+
    ret = VG_(do_syscall)(__NR_tkill, tid, signo);
-#endif /* __NR_tkill */
 
    if (ret == -VKI_ENOSYS)
       ret = VG_(do_syscall)(__NR_kill, tid, signo);
@@ -214,9 +214,9 @@ Int VG_(ktkill)( Int tid, Int signo )
    return VG_(is_kerror)(ret) ? -1 : 0;
 }
 
-Int VG_(ksigpending) ( vki_ksigset_t* set )
+Int VG_(sigpending) ( vki_sigset_t* set )
 {
-   Int res = VG_(do_syscall)(__NR_sigpending, (UInt)set);
+   Int res = VG_(do_syscall)(__NR_sigpending, (UWord)set);
    return VG_(is_kerror)(res) ? -1 : 0;
 }
 
@@ -233,8 +233,27 @@ Int VG_(gettid)(void)
 
    ret = VG_(do_syscall)(__NR_gettid);
 
-   if (ret == -VKI_ENOSYS)
-      ret = VG_(do_syscall)(__NR_getpid);
+   if (ret == -VKI_ENOSYS) {
+      Char pid[16];
+      
+      /*
+       * The gettid system call does not exist. The obvious assumption
+       * to make at this point would be that we are running on an older
+       * system where the getpid system call actually returns the ID of
+       * the current thread.
+       *
+       * Unfortunately it seems that there are some systems with a kernel
+       * where getpid has been changed to return the ID of the thread group
+       * leader but where the gettid system call has not yet been added.
+       *
+       * So instead of calling getpid here we use readlink to see where
+       * the /proc/self link is pointing...
+       */
+      if ((ret = VG_(do_syscall)(__NR_readlink, "/proc/self", pid, sizeof(pid))) >= 0) {
+         pid[ret] = '\0';
+         ret = VG_(atoll)(pid);
+      }
+   }
 
    return ret;
 }
@@ -245,24 +264,25 @@ Int VG_(gettid)(void)
    mmap/munmap, exit, fcntl
    ------------------------------------------------------------------ */
 
-static Int munmap_inner(void *start, UInt length)
+static Int munmap_inner(void *start, SizeT length)
 {
-   return VG_(do_syscall)(__NR_munmap, (UInt)start, (UInt)length );
+   return VG_(do_syscall)(__NR_munmap, (UWord)start, length );
 }
 
-static Addr mmap_inner(void *start, UInt length, UInt prot, UInt flags, UInt fd, UInt offset)
+static Addr mmap_inner(void *start, SizeT length, UInt prot, UInt flags,
+                       UInt fd, OffT offset)
 {
    Int ret;
-
-   ARCH_DO_MMAP(ret, start, ROUNDUP(length, 0x1000), prot,
-		flags & ~(VKI_MAP_NOSYMS|VKI_MAP_CLIENT),
-		fd, offset);
+   
+   PLATFORM_DO_MMAP(ret, start, length, prot,
+                    flags & ~(VKI_MAP_NOSYMS|VKI_MAP_CLIENT),
+                    fd, offset);
    return ret;
 }
 
 /* Returns -1 on failure. */
-void* VG_(mmap)( void* start, UInt length,
-                 UInt prot, UInt flags, UInt sf_flags, UInt fd, UInt offset)
+void* VG_(mmap)( void* start, SizeT length,
+                 UInt prot, UInt flags, UInt sf_flags, UInt fd, OffT offset)
 {
    Addr  res;
 
@@ -279,9 +299,9 @@ void* VG_(mmap)( void* start, UInt length,
    // Check it ended up in the right place.
    if (!VG_(is_kerror)(res)) {
       if (flags & VKI_MAP_CLIENT) {
-         vg_assert(VG_(client_base) <= res && res+length < VG_(client_end));
+         vg_assert(VG_(client_base) <= res && res+length <= VG_(client_end));
       } else {
-         vg_assert(VG_(valgrind_base) <= res && res+length-1 <= VG_(valgrind_end));
+         vg_assert(VG_(valgrind_base) <= res && res+length-1 <= VG_(valgrind_last));
       }
 
       sf_flags |= SF_MMAP;
@@ -298,7 +318,7 @@ void* VG_(mmap)( void* start, UInt length,
 }
 
 /* Returns -1 on failure. */
-Int VG_(munmap)( void* start, Int length )
+Int VG_(munmap)( void* start, SizeT length )
 {
    Int res = munmap_inner(start, length);
    if (!VG_(is_kerror)(res))
@@ -306,18 +326,29 @@ Int VG_(munmap)( void* start, Int length )
    return VG_(is_kerror)(res) ? -1 : 0;
 }
 
-Int VG_(mprotect)( void *start, Int length, UInt prot )
+Int VG_(mprotect)( void *start, SizeT length, UInt prot )
 {
-   Int res = VG_(do_syscall)(__NR_mprotect, (UInt)start, (UInt)length, prot );
+   Int res = VG_(do_syscall)(__NR_mprotect, (UWord)start, length, prot );
    if (!VG_(is_kerror)(res))
       VG_(mprotect_range)((Addr)start, length, prot);
    return VG_(is_kerror)(res) ? -1 : 0;
 }
 
+/* Terminate this single thread */
+void VG_(exit_single)( Int status )
+{
+   (void)VG_(do_syscall)(__NR_exit, status );
+   /* Why are we still alive here? */
+   /*NOTREACHED*/
+   *(volatile Int *)0 = 'x';
+   vg_assert(2+2 == 5);
+}
+
+/* Pull down the entire world */
 void VG_(exit)( Int status )
 {
-   (void)VG_(do_syscall)(__NR_exit_group, (UInt)status );
-   (void)VG_(do_syscall)(__NR_exit, (UInt)status );
+   (void)VG_(do_syscall)(__NR_exit_group, status );
+   (void)VG_(do_syscall)(__NR_exit, status );
    /* Why are we still alive here? */
    /*NOTREACHED*/
    *(volatile Int *)0 = 'x';
@@ -328,8 +359,6 @@ void VG_(exit)( Int status )
 Int VG_(fcntl) ( Int fd, Int cmd, Int arg )
 {
    Int res = VG_(do_syscall)(__NR_fcntl, fd, cmd, arg);
-   if (VG_(is_kerror)(res))
-      VG_(printf)("fcntl(%d, 0x%x, %d) gives error %d\n", fd, cmd, arg, -res);
    return VG_(is_kerror)(res) ? -1 : res;
 }
 
@@ -363,8 +392,8 @@ Int VG_(poll)( struct vki_pollfd *ufds, UInt nfds, Int timeout)
 
 /* Copy a string into the buffer. */
 static UInt
-myvprintf_str ( void(*send)(Char), Int flags, Int width, Char* str, 
-                Bool capitalise )
+myvprintf_str ( void(*send)(Char, void*), Int flags, Int width, Char* str, 
+                Bool capitalise, void *send_arg )
 {
 #  define MAYBE_TOUPPER(ch) (capitalise ? VG_(toupper)(ch) : (ch))
    UInt ret = 0;
@@ -374,14 +403,14 @@ myvprintf_str ( void(*send)(Char), Int flags, Int width, Char* str,
    if (width == 0) {
       ret += len;
       for (i = 0; i < len; i++)
-         send(MAYBE_TOUPPER(str[i]));
+          send(MAYBE_TOUPPER(str[i]), send_arg);
       return ret;
    }
 
    if (len > width) {
       ret += width;
       for (i = 0; i < width; i++)
-         send(MAYBE_TOUPPER(str[i]));
+         send(MAYBE_TOUPPER(str[i]), send_arg);
       return ret;
    }
 
@@ -389,15 +418,15 @@ myvprintf_str ( void(*send)(Char), Int flags, Int width, Char* str,
    if (flags & VG_MSG_LJUSTIFY) {
       ret += extra;
       for (i = 0; i < extra; i++)
-         send(' ');
+         send(' ', send_arg);
    }
    ret += len;
    for (i = 0; i < len; i++)
-      send(MAYBE_TOUPPER(str[i]));
+      send(MAYBE_TOUPPER(str[i]), send_arg);
    if (!(flags & VG_MSG_LJUSTIFY)) {
       ret += extra;
       for (i = 0; i < extra; i++)
-         send(' ');
+         send(' ', send_arg);
    }
 
 #  undef MAYBE_TOUPPER
@@ -412,7 +441,7 @@ myvprintf_str ( void(*send)(Char), Int flags, Int width, Char* str,
  *  WIDTH is the width of the field.
  */
 static UInt
-myvprintf_int64 ( void(*send)(Char), Int flags, Int base, Int width, ULong p)
+myvprintf_int64 ( void(*send)(Char,void*), Int flags, Int base, Int width, ULong p, void *send_arg)
 {
    Char buf[40];
    Int  ind = 0;
@@ -457,12 +486,12 @@ myvprintf_int64 ( void(*send)(Char), Int flags, Int base, Int width, ULong p)
    /* Reverse copy to buffer.  */
    ret += ind;
    for (i = ind -1; i >= 0; i--) {
-      send(buf[i]);
+      send(buf[i], send_arg);
    }
    if (width > 0 && (flags & VG_MSG_LJUSTIFY)) {
       for(; ind < width; ind++) {
 	 ret++;
-         send(' ');  // Never pad with zeroes on RHS -- changes the value!
+         send(' ', send_arg);  // Never pad with zeroes on RHS -- changes the value!
       }
    }
    return ret;
@@ -471,7 +500,7 @@ myvprintf_int64 ( void(*send)(Char), Int flags, Int base, Int width, ULong p)
 
 /* A simple vprintf().  */
 UInt
-VG_(vprintf) ( void(*send)(Char), const Char *format, va_list vargs )
+VG_(vprintf) ( void(*send)(Char,void*), const Char *format, va_list vargs, void *send_arg )
 {
    UInt ret = 0;
    int i;
@@ -486,7 +515,7 @@ VG_(vprintf) ( void(*send)(Char), const Char *format, va_list vargs )
 
    for (i = 0; format[i] != 0; i++) {
       if (format[i] != '%') {
-         send(format[i]);
+         send(format[i], send_arg);
 	 ret++;
          continue;
       }
@@ -496,7 +525,7 @@ VG_(vprintf) ( void(*send)(Char), const Char *format, va_list vargs )
          break;
       if (format[i] == '%') {
          /* `%%' is replaced by `%'. */
-         send('%');
+         send('%', send_arg);
 	 ret++;
          continue;
       }
@@ -537,42 +566,42 @@ VG_(vprintf) ( void(*send)(Char), const Char *format, va_list vargs )
             flags |= VG_MSG_SIGNED;
             if (is_long)
                ret += myvprintf_int64(send, flags, 10, width, 
-				      (ULong)(va_arg (vargs, Long)));
+				      (ULong)(va_arg (vargs, Long)), send_arg);
             else
                ret += myvprintf_int64(send, flags, 10, width, 
-				      (ULong)(va_arg (vargs, Int)));
+				      (ULong)(va_arg (vargs, Int)), send_arg);
             break;
          case 'u': /* %u */
             if (is_long)
                ret += myvprintf_int64(send, flags, 10, width, 
-				      (ULong)(va_arg (vargs, ULong)));
+				      (ULong)(va_arg (vargs, ULong)), send_arg);
             else
                ret += myvprintf_int64(send, flags, 10, width, 
-				      (ULong)(va_arg (vargs, UInt)));
+				      (ULong)(va_arg (vargs, UInt)), send_arg);
             break;
          case 'p': /* %p */
 	    ret += 2;
-            send('0');
-            send('x');
+            send('0',send_arg);
+            send('x',send_arg);
             ret += myvprintf_int64(send, flags, 16, width, 
-				   (ULong)((UInt)va_arg (vargs, void *)));
+				   (ULong)((UWord)va_arg (vargs, void *)), send_arg);
             break;
          case 'x': /* %x */
             if (is_long)
                ret += myvprintf_int64(send, flags, 16, width, 
-				      (ULong)(va_arg (vargs, ULong)));
+				      (ULong)(va_arg (vargs, ULong)), send_arg);
             else
                ret += myvprintf_int64(send, flags, 16, width, 
-				      (ULong)(va_arg (vargs, UInt)));
+				      (ULong)(va_arg (vargs, UInt)), send_arg);
             break;
          case 'c': /* %c */
 	    ret++;
-            send(va_arg (vargs, int));
+            send(va_arg (vargs, int), send_arg);
             break;
          case 's': case 'S': { /* %s */
             char *str = va_arg (vargs, char *);
             if (str == (char*) 0) str = "(null)";
-            ret += myvprintf_str(send, flags, width, str, format[i]=='S');
+            ret += myvprintf_str(send, flags, width, str, format[i]=='S', send_arg);
             break;
 	 }
 	 case 'y': { /* %y - print symbol */
@@ -588,7 +617,7 @@ VG_(vprintf) ( void(*send)(Char), const Char *format, va_list vargs )
 		  *cp++ = ')';
 		  *cp = '\0';
 	       }
-	       ret += myvprintf_str(send, flags, width, buf, 0);
+	       ret += myvprintf_str(send, flags, width, buf, 0, send_arg);
 	    }
 
 	    break;
@@ -605,35 +634,38 @@ VG_(vprintf) ( void(*send)(Char), const Char *format, va_list vargs )
    debugging info should be sent via here.  The official route is to
    to use vg_message().  This interface is deprecated.
 */
-static char myprintf_buf[100];
-static int  n_myprintf_buf;
+typedef struct {
+   char buf[100];
+   int n;
+} printf_buf;
 
-static void add_to_myprintf_buf ( Char c )
+static void add_to_myprintf_buf ( Char c, void *p )
 {
-   if (n_myprintf_buf >= 100-10 /*paranoia*/ ) {
+   printf_buf *myprintf_buf = (printf_buf *)p;
+   
+   if (myprintf_buf->n >= 100-10 /*paranoia*/ ) {
       if (VG_(clo_log_fd) >= 0) {
          VG_(send_bytes_to_logging_sink)( 
-            myprintf_buf, VG_(strlen)(myprintf_buf) );
+            myprintf_buf->buf, VG_(strlen)(myprintf_buf->buf) );
       }
-      n_myprintf_buf = 0;
-      myprintf_buf[n_myprintf_buf] = 0;      
+      myprintf_buf->n = 0;
+      myprintf_buf->buf[myprintf_buf->n] = 0;      
    }
-   myprintf_buf[n_myprintf_buf++] = c;
-   myprintf_buf[n_myprintf_buf] = 0;
+   myprintf_buf->buf[myprintf_buf->n++] = c;
+   myprintf_buf->buf[myprintf_buf->n] = 0;
 }
 
 UInt VG_(printf) ( const char *format, ... )
 {
    UInt ret;
    va_list vargs;
+   printf_buf myprintf_buf = {"",0};
    va_start(vargs,format);
    
-   n_myprintf_buf = 0;
-   myprintf_buf[n_myprintf_buf] = 0;      
-   ret = VG_(vprintf) ( add_to_myprintf_buf, format, vargs );
+   ret = VG_(vprintf) ( add_to_myprintf_buf, format, vargs, &myprintf_buf );
 
-   if (n_myprintf_buf > 0 && VG_(clo_log_fd) >= 0) {
-      VG_(send_bytes_to_logging_sink)( myprintf_buf, n_myprintf_buf );
+   if (myprintf_buf.n > 0 && VG_(clo_log_fd) >= 0) {
+      VG_(send_bytes_to_logging_sink)( myprintf_buf.buf, myprintf_buf.n );
    }
 
    va_end(vargs);
@@ -642,29 +674,27 @@ UInt VG_(printf) ( const char *format, ... )
 }
 
 /* A general replacement for sprintf(). */
-
-static Char *vg_sprintf_ptr;
-
-static void add_to_vg_sprintf_buf ( Char c )
+static void add_to_vg_sprintf_buf ( Char c, void *p )
 {
-   *vg_sprintf_ptr++ = c;
+   char **vg_sprintf_ptr = p;
+   *(*vg_sprintf_ptr)++ = c;
 }
 
 UInt VG_(sprintf) ( Char* buf, Char *format, ... )
 {
    Int ret;
    va_list vargs;
-
-   vg_sprintf_ptr = buf;
+   Char *vg_sprintf_ptr = buf;
 
    va_start(vargs,format);
 
-   ret = VG_(vprintf) ( add_to_vg_sprintf_buf, format, vargs );
-   add_to_vg_sprintf_buf(0);
+   ret = VG_(vprintf) ( add_to_vg_sprintf_buf, format, vargs, &vg_sprintf_ptr );
+   add_to_vg_sprintf_buf(0,&vg_sprintf_ptr);
 
    va_end(vargs);
 
    vg_assert(VG_(strlen)(buf) == ret);
+
    return ret;
 }
 
@@ -1032,7 +1062,8 @@ static Bool string_match_wrk ( const Char* pat, const Char* str )
    recDepth++;
    for (;;) {
       switch (*pat) {
-         case '\0' : return (*str=='\0');
+         case '\0' : recDepth--;
+                     return (*str=='\0');
          case '*'  : do {
                         if (string_match_wrk(pat+1,str)) {
                            recDepth--;
@@ -1066,6 +1097,7 @@ Bool VG_(string_match) ( const Char* pat, const Char* str )
    Bool b;
    recDepth = 0;
    b = string_match_wrk ( pat, str );
+   vg_assert(recDepth == 0);
    /*
    VG_(printf)("%s   %s   %s\n",
 	       b?"TRUE ":"FALSE", pat, str);
@@ -1087,24 +1119,27 @@ Bool VG_(string_match) ( const Char* pat, const Char* str )
 static inline ExeContext *get_real_execontext(Addr ret)
 {
    ExeContext *ec;
-   Addr esp, ebp;
-   Addr stacktop, sigstack_low, sigstack_high;
+   Addr sp, fp;
+   Addr stacktop;
+   ThreadId tid = VG_(get_lwp_tid)(VG_(gettid)());
+   ThreadState *tst = VG_(get_ThreadState)(tid);
 
-   GET_STACK_ADDRS(ebp, esp);
-   stacktop = VG_(valgrind_end);
-   VG_(get_sigstack_bounds)( &sigstack_low, &sigstack_high );
-   if (esp >= sigstack_low && esp < sigstack_high)
-      stacktop = sigstack_high;
-      
-   ec = VG_(get_ExeContext2)(ret, ebp, esp, stacktop);
+   ARCH_GET_REAL_STACK_PTR(sp);
+   ARCH_GET_REAL_FRAME_PTR(fp);
+
+   stacktop = (Addr)(tst->os_state.stack + tst->os_state.stacksize);
+
+   ec = VG_(get_ExeContext2)(ret, fp, sp, stacktop);
 
    return ec;
 }
 
 __attribute__ ((noreturn))
-static void report_and_quit ( const Char* report )
+static void report_and_quit ( const Char* report, ExeContext *ec )
 {
-   ExeContext *ec = get_real_execontext((Addr)__builtin_return_address(0));
+   if (ec == NULL)
+      ec = get_real_execontext((Addr)__builtin_return_address(0));
+
    VG_(pp_ExeContext)(ec);
    
    VG_(pp_sched_status)();
@@ -1129,7 +1164,7 @@ static void assert_fail ( const Char* expr, const Char* name, const Char* report
    entered = True;
    VG_(printf)("\n%s: %s:%d (%s): Assertion `%s' failed.\n",
                name, file, line, fn, expr );
-   report_and_quit(report);
+   report_and_quit(report, NULL);
 }
 
 void VG_(skin_assert_fail) ( const Char* expr, const Char* file, Int line, const Char* fn )
@@ -1144,21 +1179,26 @@ void VG_(core_assert_fail) ( const Char* expr, const Char* file, Int line, const
 }
 
 __attribute__ ((noreturn))
-static void panic ( Char* name, Char* report, Char* str )
+static void panic ( Char* name, Char* report, Char* str, ExeContext *ec )
 {
    VG_(printf)("\n%s: the `impossible' happened:\n   %s\n", name, str);
    VG_(printf)("Basic block ctr is approximately %llu\n", VG_(bbs_done) );
-   report_and_quit(report);
+   report_and_quit(report, ec);
 }
 
 void VG_(core_panic) ( Char* str )
 {
-   panic("valgrind", VG_BUGS_TO, str);
+   panic("valgrind", VG_BUGS_TO, str, NULL);
+}
+
+void VG_(core_panic_at) ( Char* str, ExeContext *ec )
+{
+   panic("valgrind", VG_BUGS_TO, str, ec);
 }
 
 void VG_(skin_panic) ( Char* str )
 {
-   panic(VG_(details).name, VG_(details).bug_reports_to, str);
+   panic(VG_(details).name, VG_(details).bug_reports_to, str, NULL);
 }
 
 
@@ -1186,9 +1226,6 @@ Int VG_(safe_fd)(Int oldfd)
 
    VG_(fcntl)(newfd, VKI_F_SETFD, VKI_FD_CLOEXEC);
 
-   if (newfd < VG_(fd_hard_limit))
-      VG_(printf)("oldfd=%d newfd=%d limit=%d\n",
-		  oldfd, newfd, VG_(fd_hard_limit));
    vg_assert(newfd >= VG_(fd_hard_limit));
    return newfd;
 }
@@ -1207,7 +1244,7 @@ Int VG_(open) ( const Char* pathname, Int flags, Int mode )
    /* fd = open( pathname, O_RDONLY ); */
    /* ... so we go direct to the horse's mouth, which seems to work
       ok: */
-   fd = VG_(do_syscall)(__NR_open, (UInt)pathname, flags, mode);
+   fd = VG_(do_syscall)(__NR_open, (UWord)pathname, flags, mode);
    /* VG_(printf)("result = %d\n", fd); */
    /* return -ve error code */
    return fd;
@@ -1229,7 +1266,7 @@ Int VG_(read) ( Int fd, void* buf, Int count)
 {
    Int res;
    /* res = read( fd, buf, count ); */
-   res = VG_(do_syscall)(__NR_read, fd, (UInt)buf, count);
+   res = VG_(do_syscall)(__NR_read, fd, (UWord)buf, count);
    /* return -ERRNO on error */
    return res;
 }
@@ -1238,16 +1275,16 @@ Int VG_(write) ( Int fd, const void* buf, Int count)
 {
    Int res;
    /* res = write( fd, buf, count ); */
-   res = VG_(do_syscall)(__NR_write, fd, (UInt)buf, count);
+   res = VG_(do_syscall)(__NR_write, fd, (UWord)buf, count);
    /* return -ERRNO on error */
    return res;
 }
 
-Int VG_(lseek) ( Int fd, Long offset, Int whence)
+OffT VG_(lseek) ( Int fd, OffT offset, Int whence)
 {
    Int res;
    /* res = lseek( fd, offset, whence ); */
-   res = VG_(do_syscall)(__NR_lseek, fd, (UInt)offset, whence);
+   res = VG_(do_syscall)(__NR_lseek, fd, offset, whence);
    if (VG_(is_kerror)(res)) res = -1;
    return res;
 }
@@ -1255,52 +1292,52 @@ Int VG_(lseek) ( Int fd, Long offset, Int whence)
 Int VG_(stat) ( Char* file_name, struct vki_stat* buf )
 {
    Int res;
-   res = VG_(do_syscall)(__NR_stat, (UInt)file_name, (UInt)buf);
+   res = VG_(do_syscall)(__NR_stat, (UWord)file_name, (UWord)buf);
    return res;			/* return -ve error */
 }
 
 Int VG_(fstat) ( Int fd, struct vki_stat* buf )
 {
    Int res;
-   res = VG_(do_syscall)(__NR_fstat, (UInt)fd, (UInt)buf);
+   res = VG_(do_syscall)(__NR_fstat, fd, (UWord)buf);
    return VG_(is_kerror)(res) ? (-1) : 0;
 }
 
 Int VG_(dup2) ( Int oldfd, Int newfd )
 {
    Int res;
-   res = VG_(do_syscall)(__NR_dup2, (UInt)oldfd, (UInt)newfd);
+   res = VG_(do_syscall)(__NR_dup2, oldfd, newfd);
    return VG_(is_kerror)(res) ? (-1) : res;
 }
 
 Int VG_(rename) ( Char* old_name, Char* new_name )
 {
    Int res;
-   res = VG_(do_syscall)(__NR_rename, (UInt)old_name, (UInt)new_name);
+   res = VG_(do_syscall)(__NR_rename, (UWord)old_name, (UWord)new_name);
    return VG_(is_kerror)(res) ? (-1) : 0;
 }
 
 Int VG_(unlink) ( Char* file_name )
 {
    Int res;
-   res = VG_(do_syscall)(__NR_unlink, (UInt)file_name);
+   res = VG_(do_syscall)(__NR_unlink, (UWord)file_name);
    return VG_(is_kerror)(res) ? (-1) : 0;
 }
 
 /* Nb: we do not allow the Linux extension which malloc()s memory for the
    buffer if buf==NULL, because we don't want Linux calling malloc() */
-Char* VG_(getcwd) ( Char* buf, Int size )
+Char* VG_(getcwd) ( Char* buf, SizeT size )
 {
    Int res;
    vg_assert(buf != NULL);
-   res = VG_(do_syscall)(__NR_getcwd, (UInt)buf, (UInt)size);
+   res = VG_(do_syscall)(__NR_getcwd, (UWord)buf, size);
    return VG_(is_kerror)(res) ? ((Char*)NULL) : (Char*)res;
 }
 
 /* Alternative version that does allocate the memory.  Easier to use. */
 Bool VG_(getcwd_alloc) ( Char** out )
 {
-   UInt size = 4;
+   SizeT size = 4;
 
    *out = NULL;
    while (True) {
@@ -1424,18 +1461,20 @@ Int VG_(getrlimit) (Int resource, struct vki_rlimit *rlim)
 {
    Int res;
    /* res = getrlimit( resource, rlim ); */
-   res = VG_(do_syscall)(__NR_getrlimit, (UInt)resource, (UInt)rlim);
+   res = VG_(do_syscall)(__NR_ugetrlimit, resource, (UWord)rlim);
+   if (res == -VKI_ENOSYS)
+      res = VG_(do_syscall)(__NR_getrlimit, resource, (UWord)rlim);
    if(VG_(is_kerror)(res)) res = -1;
    return res;
 }
 
 
 /* Support for setrlimit. */
-Int VG_(setrlimit) (Int resource, struct vki_rlimit *rlim)
+Int VG_(setrlimit) (Int resource, const struct vki_rlimit *rlim)
 {
    Int res;
    /* res = setrlimit( resource, rlim ); */
-   res = VG_(do_syscall)(__NR_setrlimit, (UInt)resource, (UInt)rlim);
+   res = VG_(do_syscall)(__NR_setrlimit, resource, (UWord)rlim);
    if(VG_(is_kerror)(res)) res = -1;
    return res;
 }
@@ -1446,7 +1485,7 @@ Int VG_(getdents) (UInt fd, struct vki_dirent *dirp, UInt count)
 {
    Int res;
    /* res = getdents( fd, dirp, count ); */
-   res = VG_(do_syscall)(__NR_getdents, fd, (UInt)dirp, count);
+   res = VG_(do_syscall)(__NR_getdents, fd, (UWord)dirp, count);
    if (VG_(is_kerror)(res)) res = -1;
    return res;
 }
@@ -1456,7 +1495,7 @@ Int VG_(readlink) (Char* path, Char* buf, UInt bufsiz)
 {
    Int res;
    /* res = readlink( path, buf, bufsiz ); */
-   res = VG_(do_syscall)(__NR_readlink, (UInt)path, (UInt)buf, bufsiz);
+   res = VG_(do_syscall)(__NR_readlink, (UWord)path, (UWord)buf, bufsiz);
    if (VG_(is_kerror)(res)) res = -1;
    return res;
 }
@@ -1613,7 +1652,7 @@ Int VG_(system) ( Char* cmd )
       argv[3] = 0;
 
       (void)VG_(do_syscall)(__NR_execve, 
-                            (UInt)"/bin/sh", (UInt)argv, (UInt)envp);
+                            (UWord)"/bin/sh", (UWord)argv, (UWord)envp);
 
       /* If we're still alive here, execve failed. */
       VG_(exit)(1);
@@ -1640,8 +1679,7 @@ UInt VG_(read_millisecond_timer) ( void )
    ULong now;
    Int res;
 
-   res = VG_(do_syscall)(__NR_gettimeofday, (UInt)&tv_now, 
-			 (UInt)NULL);
+   res = VG_(do_syscall)(__NR_gettimeofday, (UWord)&tv_now, (UWord)NULL);
    
    now = tv_now.tv_sec * 1000000ULL + tv_now.tv_usec;
    
@@ -1652,34 +1690,38 @@ UInt VG_(read_millisecond_timer) ( void )
 }
 
 
+void VG_(nanosleep)(struct vki_timespec *ts)
+{
+   VG_(do_syscall)(__NR_nanosleep, ts, NULL);
+}
+
 /* ---------------------------------------------------------------------
    Primitive support for bagging memory via mmap.
    ------------------------------------------------------------------ */
 
-void* VG_(get_memory_from_mmap) ( Int nBytes, Char* who )
+void* VG_(get_memory_from_mmap) ( SizeT nBytes, Char* who )
 {
-   static UInt tot_alloc = 0;
+   static SizeT tot_alloc = 0;
    void* p;
-
    p = VG_(mmap)(0, nBytes,
                  VKI_PROT_READ|VKI_PROT_WRITE|VKI_PROT_EXEC,
                  VKI_MAP_PRIVATE|VKI_MAP_ANONYMOUS, 0, -1, 0);
 
    if (p != ((void*)(-1))) {
-      vg_assert(p >= (void*)VG_(valgrind_base) && p <= (void*)VG_(valgrind_end));
-      tot_alloc += (UInt)nBytes;
+      vg_assert((void*)VG_(valgrind_base) <= p && p <= (void*)VG_(valgrind_last));
+      tot_alloc += nBytes;
       if (0)
          VG_(printf)(
-            "get_memory_from_mmap: %d tot, %d req = %p .. %p, caller %s\n",
-            tot_alloc, nBytes, p, ((char*)p) + nBytes - 1, who );
+            "get_memory_from_mmap: %llu tot, %llu req = %p .. %p, caller %s\n",
+            (ULong)tot_alloc, (ULong)nBytes, p, ((char*)p) + nBytes - 1, who );
       return p;
    }
 
    VG_(printf)("\n");
-   VG_(printf)("VG_(get_memory_from_mmap): %s's request for %d bytes failed.\n",
-               who, nBytes);
-   VG_(printf)("VG_(get_memory_from_mmap): %d bytes already allocated.\n", 
-               tot_alloc);
+   VG_(printf)("VG_(get_memory_from_mmap): %s's request for %llu bytes failed.\n",
+               who, (ULong)nBytes);
+   VG_(printf)("VG_(get_memory_from_mmap): %llu bytes already allocated.\n", 
+               (ULong)tot_alloc);
    VG_(printf)("\n");
    VG_(printf)("Sorry.  You could try using a tool that uses less memory;\n");
    VG_(printf)("eg. addrcheck instead of memcheck.\n");
@@ -1703,7 +1745,7 @@ Int VG_(log2) ( Int x )
 
 
 // Generic shell sort.  Like stdlib.h's qsort().
-void VG_(ssort)( void* base, UInt nmemb, UInt size,
+void VG_(ssort)( void* base, SizeT nmemb, SizeT size,
                  Int (*compar)(void*, void*) )
 {
    Int   incs[14] = { 1, 4, 13, 40, 121, 364, 1093, 3280,
@@ -1733,7 +1775,7 @@ void VG_(ssort)( void* base, UInt nmemb, UInt size,
    }
 
    // Specialised cases
-   if (sizeof(UInt) == size) {
+   if (sizeof(ULong) == size) {
 
       #define ASSIGN(dst, dsti, src, srci) \
       (dst)[(dsti)] = (src)[(srci)];      
@@ -1741,6 +1783,13 @@ void VG_(ssort)( void* base, UInt nmemb, UInt size,
       #define COMPAR(dst, dsti, src, srci) \
       compar( (void*)(& (dst)[(dsti)]), (void*)(& (src)[(srci)]) )
 
+      ULong* a = (ULong*)base;
+      ULong  v[1];
+
+      SORT;
+
+   } else if (sizeof(UInt) == size) {
+
       UInt* a = (UInt*)base;
       UInt  v[1];
 
@@ -1785,46 +1834,6 @@ void VG_(ssort)( void* base, UInt nmemb, UInt size,
    This is all very Linux-kernel specific.
    ------------------------------------------------------------------ */
 
-/* Various needed constants from the kernel iface (2.4),
-   /usr/src/linux-2.4.9-31 */
-
-/* kernel, ./include/linux/net.h */
-#define SYS_SOCKET      1              /* sys_socket(2)        */
-#define SYS_CONNECT     3              /* sys_connect(2)       */
-#define SYS_GETSOCKNAME 6              /* sys_getsockname(2)   */
-#define SYS_GETPEERNAME 7              /* sys_getpeername(2)   */
-#define SYS_SEND        9              /* sys_send(2)          */
-#define SYS_GETSOCKOPT  15             /* sys_getsockopt(2)    */
-
-typedef UInt __u32;
-
-/* Internet address. */
-struct vki_in_addr {
-        __u32   s_addr;
-};
-
-/* kernel, include/linux/socket.h */
-#define AF_INET                2       /* Internet IP Protocol        */
-#define MSG_NOSIGNAL           0x4000  /* Do not generate SIGPIPE */
-
-/* kernel, ./include/asm-i386/socket.h */
-#define SOCK_STREAM 1                  /* stream (connection) socket  */
-
-/* kernel, /usr/src/linux-2.4.9-31/linux/include/in.h */
-/* Structure describing an Internet (IP) socket address. */
-#define __SOCK_SIZE__   16              /* sizeof(struct sockaddr)      */
-struct vki_sockaddr_in {
-  vki_sa_family_t       sin_family;     /* Address family               */
-  unsigned short int    sin_port;       /* Port number                  */
-  struct vki_in_addr    sin_addr;       /* Internet address             */
-
-  /* Pad to size of `struct sockaddr'. */
-  unsigned char         __pad[__SOCK_SIZE__ - sizeof(short int) -
-                        sizeof(unsigned short int) - 
-                        sizeof(struct vki_in_addr)];
-};
-
-
 static
 Int parse_inet_addr_and_port ( UChar* str, UInt* ip_addr, UShort* port );
 
@@ -1877,12 +1886,12 @@ Int VG_(connect_via_socket)( UChar* str )
                   (ip >> 8) & 0xFF, ip & 0xFF, 
 	          (UInt)port );
 
-   servAddr.sin_family = AF_INET;
+   servAddr.sin_family = VKI_AF_INET;
    servAddr.sin_addr.s_addr = my_htonl(ip);
    servAddr.sin_port = my_htons(port);
 
    /* create socket */
-   sd = my_socket(AF_INET, SOCK_STREAM, 0 /* IPPROTO_IP ? */);
+   sd = my_socket(VKI_AF_INET, VKI_SOCK_STREAM, 0 /* IPPROTO_IP ? */);
    if (sd < 0) {
      /* this shouldn't happen ... nevertheless */
      return -2;
@@ -1949,11 +1958,11 @@ static
 Int my_socket ( Int domain, Int type, Int protocol )
 {
    Int res;
-   UInt args[3];
+   UWord args[3];
    args[0] = domain;
    args[1] = type;
    args[2] = protocol;
-   res = VG_(do_syscall)(__NR_socketcall, SYS_SOCKET, (UInt)&args);
+   res = VG_(do_syscall)(__NR_socketcall, VKI_SYS_SOCKET, (UWord)&args);
    if (VG_(is_kerror)(res)) 
       res = -1;
    return res;
@@ -1964,11 +1973,11 @@ Int my_connect ( Int sockfd, struct vki_sockaddr_in* serv_addr,
                  Int addrlen )
 {
    Int res;
-   UInt args[3];
+   UWord args[3];
    args[0] = sockfd;
-   args[1] = (UInt)serv_addr;
+   args[1] = (UWord)serv_addr;
    args[2] = addrlen;
-   res = VG_(do_syscall)(__NR_socketcall, SYS_CONNECT, (UInt)&args);
+   res = VG_(do_syscall)(__NR_socketcall, VKI_SYS_CONNECT, (UWord)&args);
    if (VG_(is_kerror)(res)) 
       res = -1;
    return res;
@@ -1981,15 +1990,15 @@ Int VG_(write_socket)( Int sd, void *msg, Int count )
    /* Requests not to send SIGPIPE on errors on stream oriented
       sockets when the other end breaks the connection. The EPIPE
       error is still returned. */
-   Int flags = MSG_NOSIGNAL;
+   Int flags = VKI_MSG_NOSIGNAL;
 
    Int res;
-   UInt args[4];
+   UWord args[4];
    args[0] = sd;
-   args[1] = (UInt)msg;
+   args[1] = (UWord)msg;
    args[2] = count;
    args[3] = flags;
-   res = VG_(do_syscall)(__NR_socketcall, SYS_SEND, (UInt)&args);
+   res = VG_(do_syscall)(__NR_socketcall, VKI_SYS_SEND, (UWord)&args);
    if (VG_(is_kerror)(res)) 
       res = -1;
    return res;
@@ -1998,11 +2007,11 @@ Int VG_(write_socket)( Int sd, void *msg, Int count )
 Int VG_(getsockname) ( Int sd, struct vki_sockaddr *name, Int *namelen)
 {
    Int res;
-   UInt args[3];
+   UWord args[3];
    args[0] = sd;
-   args[1] = (UInt)name;
-   args[2] = (UInt)namelen;
-   res = VG_(do_syscall)(__NR_socketcall, SYS_GETSOCKNAME, (UInt)&args);
+   args[1] = (UWord)name;
+   args[2] = (UWord)namelen;
+   res = VG_(do_syscall)(__NR_socketcall, VKI_SYS_GETSOCKNAME, (UWord)&args);
    if(VG_(is_kerror)(res))
       res = -1;
    return res;
@@ -2011,11 +2020,11 @@ Int VG_(getsockname) ( Int sd, struct vki_sockaddr *name, Int *namelen)
 Int VG_(getpeername) ( Int sd, struct vki_sockaddr *name, Int *namelen)
 {
    Int res;
-   UInt args[3];
+   UWord args[3];
    args[0] = sd;
-   args[1] = (UInt)name;
-   args[2] = (UInt)namelen;
-   res = VG_(do_syscall)(__NR_socketcall, SYS_GETPEERNAME, (UInt)&args);
+   args[1] = (UWord)name;
+   args[2] = (UWord)namelen;
+   res = VG_(do_syscall)(__NR_socketcall, VKI_SYS_GETPEERNAME, (UWord)&args);
    if(VG_(is_kerror)(res))
       res = -1;
    return res;
@@ -2025,13 +2034,13 @@ Int VG_(getsockopt) ( Int sd, Int level, Int optname, void *optval,
                       Int *optlen)
 {
    Int res;
-   UInt args[5];
+   UWord args[5];
    args[0] = sd;
-   args[1] = (UInt)level;
-   args[2] = (UInt)optname;
-   args[3] = (UInt)optval;
-   args[4] = (UInt)optlen;
-   res = VG_(do_syscall)(__NR_socketcall, SYS_GETSOCKOPT, (UInt)&args);
+   args[1] = level;
+   args[2] = optname;
+   args[3] = (UWord)optval;
+   args[4] = (UWord)optlen;
+   res = VG_(do_syscall)(__NR_socketcall, VKI_SYS_GETSOCKOPT, (UWord)&args);
    if(VG_(is_kerror)(res))
       res = -1;
    return res;
