@@ -1,5 +1,4 @@
-
-/*
+/* -*- c -*-
    ----------------------------------------------------------------
 
    Notice that the following BSD-style license applies to this one
@@ -12,7 +11,7 @@
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2004 Julian Seward.  All rights reserved.
+   Copyright (C) 2000-2005 Julian Seward.  All rights reserved.
 
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
@@ -61,6 +60,19 @@
 
 #include <stdarg.h>
 
+#undef __ppc__
+#define __ppc__   1	// Architecture we're installed on
+
+
+/* If we're not compiling for our target architecture, don't generate
+   any inline asms.  This would be a bit neater if we used the same
+   CPP symbols as the compiler for identifying architectures. */
+#if !(__x86__ && __i386__ || __powerpc__)
+#  ifndef NVALGRIND
+#    define NVALGRIND	1
+#  endif  /* NVALGRIND */
+#endif
+
 
 /* This file is for inclusion into client (your!) code.
 
@@ -70,94 +82,90 @@
    The resulting executables will still run without Valgrind, just a
    little bit more slowly than they otherwise would, but otherwise
    unchanged.  When not running on valgrind, each client request
-   consumes about 9 x86 instructions, so the resulting performance
+   consumes very few (eg. < 10) instructions, so the resulting performance
    loss is negligible unless you plan to execute client requests
    millions of times per second.  Nevertheless, if that is still a
    problem, you can compile with the NVALGRIND symbol defined (gcc
    -DNVALGRIND) so that client requests are not even compiled in.  */
 
-
-
 #ifndef NVALGRIND
-#ifdef __i386__
-/* This defines the magic code sequence which the JITter spots and
-   handles magically.  Don't look too closely at this; it will rot
-   your brain.  Valgrind dumps the result value in %EDX, so we first
-   copy the default value there, so that it is returned when not
-   running on Valgrind.  Since %EAX points to a block of mem
-   containing the args, you can pass as many args as you want like
-   this.  Currently this is set up to deal with 4 args since that's
-   the max that we appear to need (pthread_create).  
+
+/* The following defines the magic code sequences which the JITter spots and
+   handles magically.  Don't look too closely at them; they will rot
+   your brain.  We must ensure that the default value gets put in the return
+   slot, so that everything works when this is executed not under Valgrind.
+   Args are passed in a memory block, and so there's no intrinsic limit to
+   the number that could be passed, but it's currently four.
+   
+   The macro args are: 
+      _zzq_rlval    result lvalue
+      _zzq_default  default value (result returned when running on real CPU)
+      _zzq_request  request code
+      _zzq_arg1..4  request params
+
+   Nb: we put the assembly code sequences for all architectures in this one
+   file.  This is because this file must be stand-alone, so we can't rely on
+   eg. x86/ subdirectories like we do within the rest of Valgrind.
 */
-#define VALGRIND_MAGIC_SEQUENCE(                                        \
-        _zzq_rlval,   /* result lvalue */                               \
-        _zzq_default, /* result returned when running on real CPU */    \
-        _zzq_request, /* request code */                                \
-        _zzq_arg1,    /* request first param */                         \
-        _zzq_arg2,    /* request second param */                        \
-        _zzq_arg3,    /* request third param */                         \
-        _zzq_arg4     /* request fourth param */ )                      \
-                                                                        \
-  { volatile unsigned int _zzq_args[5];                                 \
-    _zzq_args[0] = (volatile unsigned int)(_zzq_request);               \
-    _zzq_args[1] = (volatile unsigned int)(_zzq_arg1);                  \
-    _zzq_args[2] = (volatile unsigned int)(_zzq_arg2);                  \
-    _zzq_args[3] = (volatile unsigned int)(_zzq_arg3);                  \
-    _zzq_args[4] = (volatile unsigned int)(_zzq_arg4);                  \
-    asm volatile("movl %1, %%eax\n\t"                                   \
-                 "movl %2, %%edx\n\t"                                   \
-                 "roll $29, %%eax ; roll $3, %%eax\n\t"                 \
-                 "rorl $27, %%eax ; rorl $5, %%eax\n\t"                 \
-                 "roll $13, %%eax ; roll $19, %%eax\n\t"                \
-                 "movl %%edx, %0\t"                                     \
-                 : "=r" (_zzq_rlval)                                    \
-                 : "r" (&_zzq_args[0]), "r" (_zzq_default)              \
-                 : "eax", "edx", "cc", "memory"                         \
-                );                                                      \
+
+#ifdef __x86__
+#define VALGRIND_MAGIC_SEQUENCE(				\
+        _zzq_rlval, _zzq_default, _zzq_request,			\
+        _zzq_arg1, _zzq_arg2, _zzq_arg3, _zzq_arg4)		\
+								\
+  { unsigned int _zzq_args[5];					\
+    _zzq_args[0] = (unsigned int)(_zzq_request);		\
+    _zzq_args[1] = (unsigned int)(_zzq_arg1);			\
+    _zzq_args[2] = (unsigned int)(_zzq_arg2);			\
+    _zzq_args[3] = (unsigned int)(_zzq_arg3);			\
+    _zzq_args[4] = (unsigned int)(_zzq_arg4);			\
+    asm volatile("roll $29, %%eax ; roll $3, %%eax\n\t"		\
+                 "rorl $27, %%eax ; rorl $5, %%eax\n\t"		\
+                 "roll $13, %%eax ; roll $19, %%eax"		\
+                 : "=d" (_zzq_rlval)				\
+                 : "a" (&_zzq_args[0]), "0" (_zzq_default)	\
+                 : "cc", "memory"				\
+                );						\
   }
-#endif
+#endif   // __x86__
+
 #ifdef __powerpc__
 #define VALGRIND_MAGIC_SEQUENCE(					\
-        _zzq_rlval,   /* result lvalue */				\
-        _zzq_default, /* result returned when running on real CPU */	\
-        _zzq_request, /* request code */				\
-        _zzq_arg1,    /* request first param */				\
-        _zzq_arg2,    /* request second param */			\
-        _zzq_arg3,    /* request third param */				\
-        _zzq_arg4     /* request fourth param */ )			\
+        _zzq_rlval, _zzq_default, _zzq_request,				\
+        _zzq_arg1, _zzq_arg2, _zzq_arg3, _zzq_arg4)			\
 									\
   { volatile unsigned int _zzq_args[5];					\
     register unsigned int _zzq_tmp __asm__("r3");			\
+    register volatile unsigned int *_zzq_ptr __asm__("r4");		\
     _zzq_args[0] = (volatile unsigned int)(_zzq_request);		\
     _zzq_args[1] = (volatile unsigned int)(_zzq_arg1);			\
     _zzq_args[2] = (volatile unsigned int)(_zzq_arg2);			\
     _zzq_args[3] = (volatile unsigned int)(_zzq_arg3);			\
     _zzq_args[4] = (volatile unsigned int)(_zzq_arg4);			\
-    _zzq_tmp = (unsigned int) _zzq_args;				\
-    asm volatile("tw 0,%0,27"						\
-		 : "=&r" (_zzq_tmp)					\
-		 : "0" (_zzq_tmp)					\
+    _zzq_ptr = _zzq_args;						\
+    asm volatile("tw 0,3,27"						\
+		 : "=r" (_zzq_tmp)					\
+		 : "0" (_zzq_default), "r" (_zzq_ptr)			\
 		 : "memory");						\
     _zzq_rlval = (__typeof__(_zzq_rlval)) _zzq_tmp;			\
   }
-#endif
+#endif   // __powerpc__
+
+// Insert assembly code for other architectures here...
+
 #else  /* NVALGRIND */
 /* Define NVALGRIND to completely remove the Valgrind magic sequence
    from the compiled code (analogous to NDEBUG's effects on
    assert())  */
 #define VALGRIND_MAGIC_SEQUENCE(					\
-        _zzq_rlval,   /* result lvalue */				\
-        _zzq_default, /* result returned when running on real CPU */	\
-        _zzq_request, /* request code */				\
-        _zzq_arg1,    /* request first param */				\
-        _zzq_arg2,    /* request second param */			\
-        _zzq_arg3,    /* request third param */				\
-        _zzq_arg4     /* request fourth param */ )			\
+        _zzq_rlval, _zzq_default, _zzq_request,                         \
+        _zzq_arg1, _zzq_arg2, _zzq_arg3, _zzq_arg4)                     \
    {									\
       (_zzq_rlval) = (_zzq_default);					\
    }
 #endif /* NVALGRIND */
 
+
 /* Some request codes.  There are many more of these, but most are not
    exposed to end-user view.  These are the public ones, all of the
    form 0x1000 + small_number.
