@@ -7,7 +7,7 @@
    This file is part of Massif, a Valgrind tool for profiling memory
    usage of programs.
 
-   Copyright (C) 2003-2004 Nicholas Nethercote
+   Copyright (C) 2003-2005 Nicholas Nethercote
       njn25@cam.ac.uk
 
    This program is free software; you can redistribute it and/or
@@ -34,7 +34,7 @@
 // be inspected step-wise to an appropriate depth.  See comments on data
 // structures below for more info on how things work.
 
-#include "vg_skin.h"
+#include "tool.h"
 //#include "vg_profile.c"
 
 #include "valgrind.h"           // For {MALLOC,FREE}LIKE_BLOCK
@@ -178,7 +178,7 @@ typedef
    struct _HP_Chunk {
       struct _HP_Chunk* next;
       Addr              data;    // Ptr to actual block
-      UInt              size;    // Size requested
+      SizeT             size;    // Size requested
       XPt*              where;   // Where allocated; bottom-XPt
    }
    HP_Chunk;
@@ -244,7 +244,7 @@ static Char buf [BUF_LEN];
 static Char buf2[BUF_LEN];
 static Char buf3[BUF_LEN];
 
-static UInt sigstacks_space = 0;    // Current signal stacks space sum
+static SizeT sigstacks_space = 0;    // Current signal stacks space sum
 
 static VgHashTable malloc_list  = NULL;   // HP_Chunks
 
@@ -265,7 +265,6 @@ static Char* alloc_fns[MAX_ALLOC_FNS] = {
    "__builtin_vec_new",
    "calloc",
    "realloc",
-   "my_malloc",   // from vg_libpthread.c
    "memalign",
 };
 
@@ -344,7 +343,7 @@ static XPt* alloc_xpt;
 
 // Cheap allocation for blocks that never need to be freed.  Saves about 10%
 // for Konqueror startup with --depth=40.
-static void* perm_malloc(UInt n_bytes)
+static void* perm_malloc(SizeT n_bytes)
 {
    static Addr hp     = 0;    // current heap pointer
    static Addr hp_lim = 0;    // maximum usable byte in current block
@@ -415,7 +414,7 @@ static Bool is_alloc_fn(Addr eip)
 // to ensure this in certain cases.  See comments below.
 static XPt* get_XCon( ThreadId tid, Bool custom_malloc )
 {
-   // Static to minimise stack size.  +1 for added 0xffffffff %eip.
+   // Static to minimise stack size.  +1 for added ~0 %eip.
    static Addr eips[MAX_DEPTH + MAX_ALLOC_FNS + 1];
 
    XPt* xpt = alloc_xpt;
@@ -440,7 +439,7 @@ static XPt* get_XCon( ThreadId tid, Bool custom_malloc )
       // wasn't a bottom-XPt (now or later) it would cause problems later (eg.
       // the parent's approx_ST wouldn't be equal [or almost equal] to the
       // total of the childrens' approx_STs).  
-      eips[ n_eips++ ] = 0xffffffff;
+      eips[ n_eips++ ] = ~((Addr)0);
 
       // Skip over alloc functions in eips[]. 
       for (L = 0; is_alloc_fn(eips[L]) && L < n_eips; L++) { }
@@ -648,7 +647,7 @@ void add_HP_Chunk(HP_Chunk* hc)
 static __inline__ 
 HP_Chunk* get_HP_Chunk(void* p, HP_Chunk*** prev_chunks_next_ptr)
 {
-   return (HP_Chunk*)VG_(HT_get_node) ( malloc_list, (UInt)p,
+   return (HP_Chunk*)VG_(HT_get_node) ( malloc_list, (UWord)p,
                                         (VgHashNode***)prev_chunks_next_ptr );
 }
 
@@ -664,7 +663,7 @@ void remove_HP_Chunk(HP_Chunk* hc, HP_Chunk** prev_chunks_next_ptr)
 static void hp_census(void);
 
 static
-void* new_block ( void* p, Int size, UInt align, Bool is_zeroed )
+void* new_block ( void* p, SizeT size, SizeT align, Bool is_zeroed )
 {
    HP_Chunk* hc;
    Bool custom_alloc = (NULL == p);
@@ -738,27 +737,27 @@ void die_block ( void* p, Bool custom_free )
 }
  
 
-void* SK_(malloc) ( Int n )
+void* SK_(malloc) ( SizeT n )
 {
    return new_block( NULL, n, VG_(clo_alignment), /*is_zeroed*/False );
 }
 
-void* SK_(__builtin_new) ( Int n )
+void* SK_(__builtin_new) ( SizeT n )
 {
    return new_block( NULL, n, VG_(clo_alignment), /*is_zeroed*/False );
 }
 
-void* SK_(__builtin_vec_new) ( Int n )
+void* SK_(__builtin_vec_new) ( SizeT n )
 {
    return new_block( NULL, n, VG_(clo_alignment), /*is_zeroed*/False );
 }
 
-void* SK_(calloc) ( Int m, Int size )
+void* SK_(calloc) ( SizeT m, SizeT size )
 {
    return new_block( NULL, m*size, VG_(clo_alignment), /*is_zeroed*/True );
 }
 
-void *SK_(memalign)( Int align, Int n )
+void *SK_(memalign)( SizeT align, SizeT n )
 {
    return new_block( NULL, n, align, False );
 }
@@ -778,13 +777,13 @@ void SK_(__builtin_vec_delete) ( void* p )
    die_block( p, /*custom_free*/False );
 }
 
-void* SK_(realloc) ( void* p_old, Int new_size )
+void* SK_(realloc) ( void* p_old, SizeT new_size )
 {
    HP_Chunk*    hc;
    HP_Chunk**   remove_handle;
    Int          i;
    void*        p_new;
-   UInt         old_size;
+   SizeT        old_size;
    XPt         *old_where, *new_where;
    
    VGP_PUSHCC(VgpCliMalloc);
@@ -1113,12 +1112,12 @@ static void hp_census(void)
 /*--- Tracked events                                       ---*/
 /*------------------------------------------------------------*/
 
-static void new_mem_stack_signal(Addr a, UInt len)
+static void new_mem_stack_signal(Addr a, SizeT len)
 {
    sigstacks_space += len;
 }
 
-static void die_mem_stack_signal(Addr a, UInt len)
+static void die_mem_stack_signal(Addr a, SizeT len)
 {
    sk_assert(sigstacks_space >= len);
    sigstacks_space -= len;
@@ -1128,13 +1127,13 @@ static void die_mem_stack_signal(Addr a, UInt len)
 /*--- Client Requests                                      ---*/
 /*------------------------------------------------------------*/
 
-Bool SK_(handle_client_request) ( ThreadId tid, UInt* argv, UInt* ret )
+Bool SK_(handle_client_request) ( ThreadId tid, UWord* argv, UWord* ret )
 {
    switch (argv[0]) {
    case VG_USERREQ__MALLOCLIKE_BLOCK: {
       void* res;
       void* p         = (void*)argv[1];
-      UInt  sizeB     =        argv[2];
+      SizeT sizeB     =        argv[2];
       *ret            = 0;
       res = new_block( p, sizeB, /*align -- ignored*/0, /*is_zeroed*/False );
       sk_assert(res == p);
@@ -1766,9 +1765,9 @@ print_summary(ULong total_ST, ULong heap_ST, ULong heap_admin_ST,
 
    // Stack(s) ----------------------------------------------------------
    if (clo_stacks) {
-      sk_assert(0 != total_ST);
       VG_(message)(Vg_UserMsg, "stack(s):          %s", 
-                   make_perc(stack_ST, total_ST) );
+                   ( 0 == stack_ST ? (Char*)"0%" 
+                                   : make_perc(stack_ST, total_ST) ) );
    }
 
    if (VG_(clo_verbosity) > 1) {
