@@ -28,16 +28,11 @@ UInt offset_value_check[5];
 
 static void load_helper_offsets(void)
 {
-    offset_value_check[0]
-	= VG_(helper_offset)((Addr) &MC_(helperc_value_check0_fail));
-    offset_value_check[1]
-	= VG_(helper_offset)((Addr) &MC_(helperc_value_check1_fail));
-    offset_value_check[2]
-	= VG_(helper_offset)((Addr) &MC_(helperc_value_check2_fail));
-    offset_value_check[3]
-	= VG_(helper_offset)((Addr) &MC_(helperc_value_check3_fail));
-    offset_value_check[4]
-	= VG_(helper_offset)((Addr) &MC_(helperc_value_check4_fail));
+    offset_value_check[0] = (Addr) &MC_(helperc_value_check0_fail);
+    offset_value_check[1] = (Addr) &MC_(helperc_value_check1_fail);
+    offset_value_check[2] = (Addr) &MC_(helperc_value_check2_fail);
+    offset_value_check[3] = (Addr) &MC_(helperc_value_check3_fail);
+    offset_value_check[4] = (Addr) &MC_(helperc_value_check4_fail);
 }
 
 static void tag1_op(TagOp op, int reg)
@@ -380,7 +375,7 @@ static void synth_loadv(int size, int reg1, int reg2)
     VG_(emit_instr)(XFORM31(reg1, 3, reg1, MO_OR, 0));
     if (dis)
 	VG_(printf)("\tmr r3,r%d\n", reg1);
-    VG_(emit_call)(VG_(helper_offset)(helper), False);
+    VG_(emit_call)(helper, False);
     VG_(emit_instr)(XFORM31(3, reg2, 3, MO_OR, 0));
     if (dis)
 	VG_(printf)("\tmr r%d,r3\n", reg2);
@@ -418,7 +413,7 @@ static void synth_storev(UInstr *u)
 	if (dis)
 	    VG_(printf)("\tmr r4,r%d\n", reg1);
     }
-    VG_(emit_call)(VG_(helper_offset)(helper), False);
+    VG_(emit_call)(helper, False);
 }
 
 unsigned int setv_sz[5] = {
