@@ -8,7 +8,7 @@
    This file is part of MemCheck, a heavyweight Valgrind tool for
    detecting memory errors.
 
-   Copyright (C) 2000-2004 Julian Seward 
+   Copyright (C) 2000-2005 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -551,10 +551,10 @@ static UCodeBlock* memcheck_instrument ( UCodeBlock* cb_in )
 
       switch (u_in->opcode) {
 
-         case LOCK:
          case NOP:
             break;
 
+         case LOCK:
          case INCEIP:
             VG_(copy_UInstr)(cb, u_in);
             break;
@@ -578,7 +578,7 @@ static UCodeBlock* memcheck_instrument ( UCodeBlock* cb_in )
 
          case GETSEG:
             sk_assert(u_in->tag2 == TempReg);
-            uInstr1(cb, SETV,  2, TempReg, SHADOW(u_in->val2));
+            uInstr1(cb, SETV,  u_in->size, TempReg, SHADOW(u_in->val2));
             VG_(copy_UInstr)(cb, u_in);
             break;
 
